"""
__author__ = "David Frank"
__copyright__ = "Copyright 2025, Pomfort GmbH"
__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from os.path import abspath
from pathlib import Path

from click.testing import CliRunner
from freezegun import freeze_time
import ascmhl.commands
from ascmhl.generator import _extract_pattern_relative_to_history
from ascmhl.history import MHLHistory
from tests.conftest import abspath_conversion_tests, remove_tree
from tests.test_ignore import assert_mhl_file_has_exact_ignore_patterns, assert_pattern_ignored_in_result
from ascmhl.generator import belongs_to_parent_or_neighbour


@freeze_time("2020-01-16 09:15:00")
def test_ignore_relative_path_pattern_1(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/A/A1.txt"})
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_relative_path_pattern_2(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "/A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/A/A1.txt"})
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_relative_path_pattern_3(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "**/A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/**/A/A1.txt"})
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_1(fs, nested_mhl_histories):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64", "-i", "/AA/AA1.txt", "-v"]
    )
    print(result.output)
    assert_mhl_file_has_exact_ignore_patterns("root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl", {"/AA1.txt"})
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/A")).latest_ignore_patterns()
    assert "/AA/AA1.txt" not in latest_ignores

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root")).latest_ignore_patterns()
    assert "/AA/AA1.txt" not in latest_ignores


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_2(fs, nested_mhl_histories):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64", "-i", "AA/AA1.txt", "-v"]
    )
    print(result.output)
    assert_mhl_file_has_exact_ignore_patterns("root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl", {"/AA1.txt"})
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/A")).latest_ignore_patterns()
    assert "AA/AA1.txt" not in latest_ignores

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root")).latest_ignore_patterns()
    assert "AA/AA1.txt" not in latest_ignores


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_3(fs, nested_mhl_histories):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64", "-i", "**/AA1.txt", "-v"]
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    assert_mhl_file_has_exact_ignore_patterns("root/A/ascmhl/0001_A_2020-01-16_091500Z.mhl", {"/**/AA1.txt"})
    assert_mhl_file_has_exact_ignore_patterns("root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl", {"/**/AA1.txt"})

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_4(fs, nested_mhl_histories):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "AA1.txt", "-v"]
    )
    print(result.output)
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"AA1.txt"})
    assert_mhl_file_has_exact_ignore_patterns("root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl", {"AA1.txt"})
    assert_mhl_file_has_exact_ignore_patterns("root/B/BB/ascmhl/0002_BB_2020-01-16_091500Z.mhl", {"AA1.txt"})
    assert_mhl_file_has_exact_ignore_patterns("root/B/ascmhl/0002_B_2020-01-16_091500Z.mhl", {"AA1.txt"})
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_5(fs, post_house_file_structure):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/ShootingDay1"), "-h", "xxh64", "-i", "**/A/A001", "-v"]
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/A/A001"], result)
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/ascmhl/0002_ShootingDay1_2020-01-16_091500Z.mhl", {"/**/A/A001"}
    )
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/CameraMedia/A/ascmhl/0004_A_2020-01-16_091500Z.mhl", {"/**/A/A001"}
    )
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/CameraMedia/B/ascmhl/0004_B_2020-01-16_091500Z.mhl", {"/**/A/A001"}
    )


@freeze_time("2020-01-16 09:15:00")
def test_ignore_nested_path_pattern_6(fs, post_house_file_structure):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-h", "xxh64", "-i", "**/?/Proxy/", "-v"],
    )
    print(result.output)
    assert result.exit_code == 0
    pattern = ["/root/ShootingDay1/CameraMedia/A/Proxy", "/root/ShootingDay1/CameraMedia/B/Proxy"]
    assert_pattern_ignored_in_result(pattern, result)
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/ascmhl/0002_ShootingDay1_2020-01-16_091500Z.mhl", {"/**/?/Proxy/"}
    )
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/CameraMedia/A/ascmhl/0004_A_2020-01-16_091500Z.mhl", {"/**/?/Proxy/"}
    )
    assert_mhl_file_has_exact_ignore_patterns(
        "root/ShootingDay1/CameraMedia/B/ascmhl/0004_B_2020-01-16_091500Z.mhl", {"/**/?/Proxy/"}
    )


@freeze_time("2020-01-16 09:15:00")
def test_ignore_future_files(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "A/A2.txt", "-v"]
    )
    assert result.exit_code == 0

    fs.create_file("/root/A/A2.txt")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A2.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_existing_files(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/A/A1.txt"})
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_deleted_files(fs, simple_mhl_history):
    runner = CliRunner()
    os.remove("/root/A/A1.txt")
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/A/A1.txt"})
    print(result.output)
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_ignore_changing_files(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "A/A1.txt", "-v"]
    )
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0002_root_2020-01-16_091500Z.mhl", {"/A/A1.txt"})
    print(result.output)

    os.remove("/root/A/A1.txt")
    fs.create_file("/root/A/A1.txt", contents="1234567890")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A1.txt"], result)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_in_first_generation(fs):
    runner = CliRunner()

    fs.create_file("/root/A/A1.txt", contents="1234567890")
    fs.create_file("/root/A/A1.RMD", contents="1234567890")
    fs.create_file("/root/A/A2.txt", contents="1234567890")
    fs.create_file("/root/A/A2.RMD", contents="1234567890")
    fs.create_file("/root/A/ignore.ign", contents="1234567890")
    fs.create_file("/root/report.pdf", contents="1234567890")
    fs.create_file("/root/report.txt", contents="1234567890")
    fs.create_file("/root/ignore.ign", contents="1234567890")

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-i", "*.txt", "-i", "ignore.ign"],
    )
    print(result.output)
    assert result.exit_code == 0
    pattern = ["/root/A/A1.txt", "/root/A/A2.txt", "/root/A/ignore.ign", "/root/ignore.ign", "/root/report.txt"]
    assert_pattern_ignored_in_result(pattern, result)
    assert_mhl_file_has_exact_ignore_patterns("root/ascmhl/0001_root_2020-01-16_091500Z.mhl", {"*.txt", "ignore.ign"})


def test_ignore_paths_are_handled_correctly(fs, nested_mhl_histories):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64", "-v", "-i", "/AB/AB1.txt"]
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AB/AB1.txt"], result)

    fs.create_file("/root/AB/AB1.txt")
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-i", "Stuff.txt"]
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AB/AB1.txt", "/root/Stuff.txt"], result)

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v", "-i", "A/AA/AA1.txt"]
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)

    os.remove("/root/A/AA/AA1.txt")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0

    os.remove("/root/A/AB/AB1.txt")
    os.remove("/root/Stuff.txt")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0

    os.remove("/root/B/B1.txt")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 10


def test_nested_histories_absolute_ignore_patterns(fs, nested_mhl_histories):
    runner = CliRunner()

    # existing histories: /root, /root/A/AA, /root/B, /root/B/BB (1st gen)
    # create a history with an ignore pattern
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-i", "/test.txt"]
    )
    assert result.exit_code == 0

    # create two test files
    fs.create_file("/root/A/AA/test.txt", contents="testAtSubHistoryRoot")
    fs.create_file("/root/A/AA/Subfolder/test.txt", contents="testInSubHistorySubfolder")

    # run the create once in the sub history and see if the ignore pattern works
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0
    hash_list = MHLHistory.load_from_path(abspath("/root/A/AA")).hash_lists[-1]

    # the ignore pattern /test.txt should ignore this file
    assert hash_list.find_media_hash_for_path("test.txt") is None

    # the ignore pattern /test.txt should not ignore this file
    assert hash_list.find_media_hash_for_path(f"{Path('Subfolder/test.txt')}") is not None

    # same ignores should be applied when running it on the root history, so we run it again on the root hsitory
    result = runner.invoke(ascmhl.commands.create, [abspath("/root"), "-h", "xxh64", "-v"])
    assert result.exit_code == 0
    hash_list = MHLHistory.load_from_path(abspath("/root/A/AA")).hash_lists[-1]

    # the ignore pattern /test.txt from the sub history should ignore this file also when verifying from the root
    assert hash_list.find_media_hash_for_path("test.txt") is None

    # the ignore pattern /test.txt from the sub history should not ignore this file
    assert hash_list.find_media_hash_for_path(f"{Path('Subfolder/test.txt')}") is not None


def test_nested_histories_ignoring_from_root(fs, nested_mhl_histories):
    runner = CliRunner()

    # existing histories: /root, /root/A/AA, /root/B, /root/B/BB (1st gen)
    # create two test files
    fs.create_file("/root/A/AA/test.txt", contents="testAtSubHistoryRoot")
    fs.create_file("/root/A/AA/Subfolder/test.txt", contents="testInSubHistorySubfolder")

    # ignore a file in the subhistory from the root hsitory
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-i", "/A/AA/test.txt"]
    )
    assert result.exit_code == 0

    hash_list = MHLHistory.load_from_path(abspath("/root/A/AA")).hash_lists[-1]

    # the ignore pattern should ignore this file
    assert hash_list.find_media_hash_for_path("test.txt") is None

    # the ignore pattern should also ignore this file
    assert hash_list.find_media_hash_for_path(f"{Path('Subfolder/test.txt')}") is not None

    # the ignore pattern inside the sub history should contain the pattern /test.txt not the path we passed into the root history
    latest_ignores = MHLHistory.load_from_path(abspath("/root/A/AA")).latest_ignore_patterns()
    assert "/test.txt" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root")).latest_ignore_patterns()
    assert "/test.txt" not in latest_ignores

    # a second verification should not fail
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert result.exit_code == 0


def test_ignore_multilevel_histories(fs, nested_mhl_histories):
    runner = CliRunner()

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64"])
    assert result.exit_code == 0

    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A/AB"), "-v", "-h", "xxh64", "-i", "AB1.txt"]
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AB/AB1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/A/AB")).latest_ignore_patterns()
    assert "AB1.txt" in latest_ignores

    assert result.exit_code == 0
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B/BA"), "-h", "xxh64"])
    assert result.exit_code == 0
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64"])
    assert result.exit_code == 0

    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64", "-i", "/A/AA/AA1.txt"]
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA1.txt"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/A/AA")).latest_ignore_patterns()
    assert "/AA1.txt" in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/A")).latest_ignore_patterns()
    assert "/AA/AA1.txt" not in latest_ignores
    assert "/AA1.txt" not in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/B")).latest_ignore_patterns()
    assert "/AA/AA1.txt" not in latest_ignores
    assert "/AA1.txt" not in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root")).latest_ignore_patterns()
    assert "/A/AA/AA1.txt" not in latest_ignores
    assert "/AA/AA1.txt" not in latest_ignores
    assert "/AA1.txt" not in latest_ignores


def test_ignore_files_in_nested_structures(fs, post_house_file_structure):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "Sidecar.txt", "-h", "xxh64"],
    )
    assert result.exit_code == 0

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/Sound")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores

    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt",
        "/root/ShootingDay1/Sound/Sidecar.txt",
    ]
    assert_pattern_ignored_in_result(pattern, result)


def test_ignore_file_type_in_nested_histories(fs, post_house_file_structure):
    """
    This should test the '*' functionality
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root"), "-v", "-i", "*.pdf", "-h", "xxh64"],
    )
    assert result.exit_code == 0

    assert_pattern_ignored_in_result(
        ["/root/ShootingDay1/CameraMedia/Report.pdf", "/root/ShootingDay1/Report.pdf"], result
    )

    latest_ignores = MHLHistory.load_from_path(abspath("/root")).latest_ignore_patterns()
    assert "*.pdf" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "*.pdf" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/Sound")).latest_ignore_patterns()
    assert "*.pdf" in latest_ignores

    fs.create_file("/root/ShootingDay1/Sound/Notes.pdf")
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root"), "-v", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/ShootingDay1/Sound/Notes.pdf"], result)


def test_ignore_file_type_in_path(fs, post_house_file_structure_with_range):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "CameraMedia/A/A002/*.mov", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/A002/*.mov" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A002/A002C001.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C002.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C003.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C004.mov",
    ]
    assert_pattern_ignored_in_result(pattern, result)

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "/CameraMedia/A/A002/*.mov" not in latest_ignores
    assert "CameraMedia/A/A002/*.mov" not in latest_ignores

    # verify it again without providing the ignore pattern
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/A002/*.mov" in latest_ignores
    assert_pattern_ignored_in_result(pattern, result)


def test_ignore_directories_in_nested_structures_pattern_1(fs, post_house_file_structure):
    """
    This pattern should match "Proxy" anywhere in the filepath
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/CameraMedia"), "-v", "-i", "Proxy", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "Proxy" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
        "/root/ShootingDay1/CameraMedia/Proxy",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    assert result.exit_code == 0


def test_ignore_directories_in_nested_structures_pattern_2(fs, post_house_file_structure):
    """
    This pattern should only match directories named "Proxy" anywhere in the directory tree, but not files with that name
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/CameraMedia"), "-v", "-i", "Proxy/", "-h", "xxh64"],
    )
    pattern = ["/root/ShootingDay1/CameraMedia/A/Proxy", "/root/ShootingDay1/CameraMedia/B/Proxy"]
    assert_pattern_ignored_in_result(pattern, result)
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/Proxy"], result, negate=True)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "Proxy/" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "Proxy/" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/B")).latest_ignore_patterns()
    assert "Proxy/" in latest_ignores
    assert result.exit_code == 0


def test_ignore_directories_in_nested_structures_pattern_3(fs, post_house_file_structure):
    """
    This pattern should only match /A/Proxy/ relative to CameraMedia and not be applied to other directories
    """
    runner = CliRunner()
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/A/Proxy/B001/A3_p.txt", contents="A3 Proxy")

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/CameraMedia"), "-v", "-i", "A/Proxy/", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/Proxy/" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/B")).latest_ignore_patterns()
    assert "A/Proxy/" not in latest_ignores
    assert "/A/Proxy/" not in latest_ignores
    assert "//A/Proxy/" not in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "A/Proxy/" not in latest_ignores
    assert "/A/Proxy/" not in latest_ignores
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/A/Proxy"], result)
    assert_pattern_ignored_in_result(
        ["/root/ShootingDay1/CameraMedia/B/Proxy", "/root/ShootingDay1/CameraMedia/Proxy"], result, negate=True
    )
    assert f"created original hash for     {Path('B/Proxy/A/Proxy/B001/A3_p.txt')}" in result.output
    assert result.exit_code == 0


def test_ignore_directories_in_nested_structures_pattern_3_wrong_directory(fs, post_house_file_structure):
    """
    This pattern should only match /A/Proxy/ relative to the ascmhl and not be applied to other directories
    (i.e. have no effect when called on root/ShootingDay1, since it does not contain a directory 'A/Proxy/')
    """
    runner = CliRunner()
    fs.create_file("/root/ShootingDay1/CameraMedia/B/Proxy/A/Proxy/B001/A3_p.txt", contents="A3 Proxy")

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "A/Proxy/", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/Proxy/" not in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/B")).latest_ignore_patterns()
    assert "/A/Proxy/" not in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/A/Proxy/" not in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
        "/root/ShootingDay1/CameraMedia/Proxy",
    ]
    assert_pattern_ignored_in_result(pattern, result, negate=True)
    assert f"created original hash for     {Path('CameraMedia/B/Proxy/A/Proxy/B001/A3_p.txt')}" in result.output
    assert result.exit_code == 0


def test_ignore__directories_in_nested_structures_pattern_4(fs, post_house_file_structure):
    """
    This pattern should ignore any occurence of Proxy in the file- or directoryname below root/ShootingDay1
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "**/Proxy", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "/**/Proxy" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/**/Proxy" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/**/Proxy" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
        "/root/ShootingDay1/CameraMedia/Proxy",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    assert result.exit_code == 0


def test_ignore__directories_in_nested_structures_pattern_5(fs, post_house_file_structure):
    """
    This pattern should ignore any directory 'Proxy', but not a file named 'Proxy'
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "**/Proxy/", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "/**/Proxy/" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/**/Proxy/" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/**/Proxy/" in latest_ignores
    pattern = ["/root/ShootingDay1/CameraMedia/A/Proxy", "/root/ShootingDay1/CameraMedia/B/Proxy"]
    assert_pattern_ignored_in_result(pattern, result)
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/Proxy"], result, negate=True)
    assert result.exit_code == 0


def test_ignore__directories_in_nested_structures_pattern_6(fs, post_house_file_structure):
    """
    This pattern should ignore any files in any directory 'Proxy', but not a file named 'Proxy'
    """
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "**/Proxy/**", "-h", "xxh64"],
    )
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "**/Proxy/**" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "**/Proxy/**" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "**/Proxy/**" in latest_ignores
    pattern = ["/root/ShootingDay1/CameraMedia/A/Proxy", "/root/ShootingDay1/CameraMedia/B/Proxy"]
    assert_pattern_ignored_in_result(pattern, result)
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/Proxy"], result, negate=True)
    assert result.exit_code == 0


def test_ignore_single_character(fs, post_house_file_structure_with_range):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/CameraMedia/A"), "-v", "-i", "A002C00?.mov", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "A002C00?.mov" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A002/A002C001.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C002.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C003.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C004.mov",
    ]
    assert_pattern_ignored_in_result(pattern, result)


def test_ignore_single_character_in_path(fs, post_house_file_structure_with_range):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "CameraMedia/A/A002/A002C00?.mov", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/A002/A002C00?.mov" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A002/A002C001.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C002.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C003.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C004.mov",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "CameraMedia/A/A002C00?.mov" not in latest_ignores


def test_ignore_multiple_individual_characters_in_path(fs, post_house_file_structure):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [
            abspath_conversion_tests("/root/ShootingDay1"),
            "-v",
            "-i",
            "CameraMedia/?/Proxy/?001/?001.ale",
            "-h",
            "xxh64",
        ],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/?/Proxy/?001/?001.ale" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/Proxy/A001/A001.ale",
        "/root/ShootingDay1/CameraMedia/B/Proxy/B001/B001.ale",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "CameraMedia/?/Proxy/?001/?001.ale" not in latest_ignores


def test_ignore_range_of_characters(fs, post_house_file_structure_with_range):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [
            abspath_conversion_tests("/root/ShootingDay1"),
            "-v",
            "-i",
            "CameraMedia/A/A00[1-5]/A00?C00[1-9].mov",
            "-h",
            "xxh64",
        ],
    )
    assert result.exit_code == 0
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "/A00[1-5]/A00?C00[1-9].mov" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A002/A002C001.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C002.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C003.mov",
        "/root/ShootingDay1/CameraMedia/A/A002/A002C004.mov",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "CameraMedia/A/A00[1-5]/A00?C00[1-9].mov" not in latest_ignores


def test_ignore_negate_pattern(fs, post_house_file_structure):
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "Sidecar.txt", "-h", "xxh64"],
    )
    assert result.exit_code == 0

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/Sound")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt",
        "/root/ShootingDay1/Sound/Sidecar.txt",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/Sound"), "-v", "-i", "!Sidecar.txt", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/ShootingDay1/Sound/Sidecar.txt"], result, negate=True)


@freeze_time("2020-01-16 09:15:00")
def test_ignore_old_deleted_files_in_histories(fs, nested_mhl_histories):
    runner = CliRunner()

    # existing histories: /root, /root/A/AA, /root/B, /root/B/BB (1st gen)

    fs.create_file("/root/A/AA/AA1.RMD", contents="Lorem ipsum dolor")
    fs.create_file("/root/A/AB/AB1.RMD", contents="sit amet con vota")
    fs.create_file("/root/B/BA/B1.RMD", contents="lirum alamru aexti")
    fs.create_file("/root/A/AA/AA2.txt", contents="Lorem ipsum dolor")

    result = runner.invoke(ascmhl.commands.create, [abspath("/root"), "-h", "xxh64"])
    assert result.exit_code == 0
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/A/AA"), "-h", "xxh64"])
    assert result.exit_code == 0
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/B"), "-h", "xxh64"])
    assert result.exit_code == 0
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/B/BB"), "-h", "xxh64"])
    assert result.exit_code == 0

    # existing histories: /root, /root/A/AA, /root/B, /root/B/BB (2nd gen)

    print("First session, no ignores yet\n")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-h", "xxh64", "-v"])
    print(result.output)
    assert result.exit_code == 0

    print("Second Session, should ignore single existing file AB1.txt in root/A/AB/AB1.txt\n")

    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/A"), "-v", "-h", "xxh64", "-i", "AB/AB1.txt"]
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AB/AB1.txt"], result)
    assert_mhl_file_has_exact_ignore_patterns("root/A/ascmhl/0002_A_2020-01-16_091500Z.mhl", {"/AB/AB1.txt"})

    print('Third Session, ignore "*.txt" and "/root/A/AB/AB1.RMD "\nRemove root/A/AA/AA1.txt and /root/A/AB/AB1.txt\n')
    os.remove("/root/A/AA/AA1.txt")
    os.remove("/root/A/AB/AB1.txt")
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/A"), "-v", "-h", "xxh64", "-i", "*.txt", "-i", "AB/AB1.RMD"],
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/AA/AA2.txt", "/root/A/AB/AB1.RMD"], result)
    assert_mhl_file_has_exact_ignore_patterns(
        "root/A/ascmhl/0003_A_2020-01-16_091500Z.mhl", {"*.txt", "/AB/AB1.RMD", "/AB/AB1.txt"}
    )

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64"])
    print(result.output)
    assert result.exit_code == 0

    os.remove("/root/B/BB/BB1.txt")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v", "-h", "xxh64"])
    print(result.output)
    assert result.exit_code == 10


def test_ignore_multiple_pattern_in_single_command(fs, post_house_file_structure):
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/Report.pdf", contents="A1")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/ClipList.pdf", contents="A1")
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [
            abspath_conversion_tests("/root/ShootingDay1"),
            "-v",
            "-h",
            "xxh64",
            "-i",
            "Sidecar.txt",
            "-i",
            "Proxy/",
            "-i",
            "CameraMedia/**/*.pdf",
            "-i",
            "/CameraMedia/B/B002/B3.txt",
        ],
    )
    assert result.exit_code == 0

    pattern = [
        "/root/ShootingDay1/CameraMedia/Report.pdf",
        "/root/ShootingDay1/CameraMedia/A/A001/Report.pdf",
        "/root/ShootingDay1/CameraMedia/B/B002/ClipList.pdf",
        "/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt",
        "/root/ShootingDay1/Sound/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "Proxy/" in latest_ignores
    assert "/CameraMedia/**/*.pdf" not in latest_ignores
    assert "/CameraMedia/B/B002/B3.txt" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/Sound")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "Proxy/" in latest_ignores
    assert "CameraMedia/**/*.pdf" not in latest_ignores
    assert "/CameraMedia/B/B002/B3.txt" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "Proxy/" in latest_ignores
    assert "/**/*.pdf" in latest_ignores
    assert "/CameraMedia/B/B002/B3.txt" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "Proxy/" in latest_ignores
    assert "/**/*.pdf" in latest_ignores
    assert "/B002/B3.txt" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/B")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "Proxy/" in latest_ignores
    assert "/**/*.pdf" in latest_ignores
    assert "/B002/B3.txt" in latest_ignores


def test_ignore_diff_nested_multiple_pattern(fs, post_house_file_structure):
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/Report.pdf", contents="A1")
    fs.create_file("/root/ShootingDay1/CameraMedia/B/B002/ClipList.pdf", contents="A1")
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.diff,
        [
            abspath_conversion_tests("/root/ShootingDay1"),
            "-v",
            "-i",
            "Sidecar.txt",
            "-i",
            "Proxy/",
            "-i",
            "CameraMedia/**/*.pdf",
            "-i",
            "/CameraMedia/B/B002/B3.txt",
            "-i",
            "/CameraMedia/Proxy",
        ],
    )
    assert result.exit_code == 0

    pattern = [
        "/root/ShootingDay1/CameraMedia/Report.pdf",
        "/root/ShootingDay1/CameraMedia/A/A001/Report.pdf",
        "/root/ShootingDay1/CameraMedia/B/B002/ClipList.pdf",
        "/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt",
        "/root/ShootingDay1/Sound/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
    ]
    assert_pattern_ignored_in_result(pattern, result)


def test_ignore_from_file(fs, post_house_file_structure):
    fs.create_file(
        "/user/ignore.txt", contents="Sidecar.txt\n/CameraMedia/**/Proxy/**\nCameraMedia/A/A001/A1.txt\n/**/temp\n/dir"
    )

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-h", "xxh64", "-ii", "/user/ignore.txt"],
    )
    assert result.exit_code == 0

    pattern = [
        "/root/ShootingDay1/CameraMedia/A/A001/A1.txt",
        "/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/B/B002/Sidecar.txt",
        "/root/ShootingDay1/CameraMedia/A/Proxy",
        "/root/ShootingDay1/CameraMedia/B/Proxy",
        "/root/ShootingDay1/Sound/Sidecar.txt",
    ]
    assert_pattern_ignored_in_result(pattern, result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/A")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "/A001/A1.txt" in latest_ignores
    assert "/**/temp" in latest_ignores
    assert "/dir" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia/B")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "/A001/A1.txt" not in latest_ignores
    assert "/**/temp" in latest_ignores
    assert "/dir" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/**/Proxy/**" in latest_ignores
    assert "/A001/A1.txt" not in latest_ignores
    assert "/**/temp" in latest_ignores
    assert "/dir" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/Sound")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "/CameraMedia/**/Proxy/**" not in latest_ignores
    assert "/CameraMedia/A001/A1.txt" not in latest_ignores
    assert "/**/temp" in latest_ignores
    assert "/dir" not in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "Sidecar.txt" in latest_ignores
    assert "/CameraMedia/**/Proxy/**" not in latest_ignores
    assert "/CameraMedia/A001/A1.txt" not in latest_ignores
    assert "/**/temp" in latest_ignores
    assert "/dir" in latest_ignores


def test_ignore_history_in_ignored_directory(fs):
    runner = CliRunner()
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A1.txt", contents="A1")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A2.txt", contents="A2")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/A3.txt", contents="A3")
    fs.create_file("/root/ShootingDay1/CameraMedia/A/A001/Sidecar.txt", contents="Sidecar1")
    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1"), "-h", "xxh64"])
    assert result.exit_code == 0

    result = runner.invoke(ascmhl.commands.create, [abspath("/root/ShootingDay1/CameraMedia"), "-h", "xxh64"])
    assert result.exit_code == 0

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1/CameraMedia/A/A001"), "-v", "-h", "xxh64"],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-i", "/CameraMedia/A/A001", "-h", "xxh64"],
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/A/A001"], result)
    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1/CameraMedia")).latest_ignore_patterns()
    assert "/A/A001" in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/ShootingDay1")).latest_ignore_patterns()
    assert "/CameraMedia/A/A001" not in latest_ignores

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/ShootingDay1"), "-v", "-h", "xxh64"],
    )
    print(result.output)
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/ShootingDay1/CameraMedia/A/A001"], result)


def test_ignore_only_one_child_history(fs):
    fs.create_file("/root/A/A001/A001C001.txt")
    fs.create_file("/root/A/A001/A001C002.txt")
    fs.create_file("/root/A/A001/A001C003.txt")

    fs.create_file("/root/A/A002/A002C001.txt")
    fs.create_file("/root/A/A002/A002C002.txt")
    fs.create_file("/root/A/A002/A002C003.txt")

    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/A/A002"), "-v", "-h", "xxh64"],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        ascmhl.commands.create,
        [abspath_conversion_tests("/root/A"), "-v", "-i", "/A001/A001C001.txt", "-h", "xxh64"],
    )
    assert result.exit_code == 0
    assert_pattern_ignored_in_result(["/root/A/A001/A001C001.txt"], result)

    latest_ignores = MHLHistory.load_from_path(abspath("/root/A")).latest_ignore_patterns()
    assert "/A001/A001C001.txt" in latest_ignores

    latest_ignores = MHLHistory.load_from_path(abspath("/root/A/A002")).latest_ignore_patterns()
    assert "/A001/A001C001.txt" not in latest_ignores


def test_ignore_unit_belongs_to_parent_or_neighbour():
    pattern = "*.txt"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "A.txt"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "A/"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "/A"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/AA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/AA/"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/AA/AAA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/AA/AAA.txt"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/A/AA/AAA/A.txt"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "A/**/AA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "A/AA/**"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "**/AA/"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == False

    pattern = "/**/AA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/**/AA/"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "A/B/AAA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "A/AA/AAB"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "A/AA/AAB/AA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "B/AA/AAA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True

    pattern = "/B/AA/AAA"
    parent_rel_path = f"{Path('/A/AA/AAA')}"
    result = ascmhl.generator.belongs_to_parent_or_neighbour(pattern, parent_rel_path)
    assert result == True


def test_ignore_unit_extract_relpath():
    base = f"{Path('/home/user/docs')}"
    target = "/home/user/docs/file.txt"
    expected = "file.txt"
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/home/user')}"
    target = "/home/user/projects/code"
    expected = "projects/code"
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/home/user/docs')}"
    target = "/home/user"
    expected = None
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/home/user/docs')}"
    target = "/home/user/photos/image.jpg"
    expected = "photos/image.jpg"
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/home/user/docs')}"
    target = "/home/user/docs"
    expected = None
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/')}"
    target = "/var/log"
    expected = "var/log"
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected

    base = f"{Path('/home/user')}"
    target = "/home/user/docs/"
    expected = "docs/"
    result = ascmhl.generator._extract_pattern_relative_to_history(target, base)
    assert result == expected


def test_ignore_missing_subhistory(fs, post_house_file_structure):
    """
    test that creating a new generation doesn't fail if we ignore a missing subhistory
    """

    # delete the whole subhistory, it is referenced from the history in the CameraMedia folder but as we ignore it
    # the create command doesn't fail
    remove_tree("/root/ShootingDay1/CameraMedia/A")
    runner = CliRunner()
    result = runner.invoke(
        ascmhl.commands.create, [abspath_conversion_tests("/root/ShootingDay1/CameraMedia"), "-i", "/A"]
    )
    assert result.exit_code == 0



"""
__author__ = "Patrick Renner, Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from __future__ import annotations
from typing import List, Dict, Optional, Set
from datetime import datetime
import os

from . import logger
from .ignore import MHLIgnoreSpec
from .__version__ import ascmhl_reference_hash_format
from .hasher import hash_file


class MHLHashList:
    """
    class for representing one MHL generation
    managed by a MHLHistory object
    uses MHLMediaHash class for storing information about one file

    - public interface
        * accessing derived information (e.g. summaries)

    - private interface
        * initialize new, empty MHLHash for adding one files with hash(es)

    model member variables:
    history -- MHLHistory object for context
    media_hashes -- list of MHLMediaHash objects

    attribute member variables:
    generation_number -- generation number of the represented MHL file

    other member variables:
    file path -- file path of the represented MHL file
    """

    creator_info: Optional[MHLCreatorInfo]
    process_info: MHLProcessInfo
    media_hashes: List[MHLMediaHash]
    media_hashes_path_map: Dict[str, MHLMediaHash]
    # referenced_hash_lists are the loaded hash list object
    referenced_hash_lists = List["MHLHashList"]
    # while hash_list_references store the reference objects found in the mhl files
    hash_list_references = List["MHLHashListReference"]
    file_path: Optional[str]
    generation_number: Optional[int]

    def __init__(self):
        self.creator_info = None
        self.process_info = MHLProcessInfo()
        self.media_hashes = []
        self.file_path = None
        self.generation_number = None
        self.referenced_hash_lists = []
        self.hash_list_references = []
        self.media_hashes_path_map = {}

    # methods to query for hashes
    def find_media_hash_for_path(self, relative_path):
        return self.media_hashes_path_map.get(relative_path)

    def find_or_create_media_hash_for_path(self, relative_path, file_size, file_modification_date):
        media_hash = self.find_media_hash_for_path(relative_path)
        if not media_hash:
            media_hash = MHLMediaHash()
            media_hash.path = relative_path
            media_hash.file_size = file_size
            media_hash.last_modification_date = file_modification_date
            self.append_hash(media_hash)
        return media_hash

    def set_of_file_paths(self, root_path) -> Set[str]:
        all_paths = set()
        for media_hash in self.media_hashes:
            all_paths.add(os.path.join(root_path, media_hash.path))
        return all_paths

    def renamed_path_with_previous_path(self, root_path):
        all_paths = {}
        for media_hash in self.media_hashes:
            if media_hash.previous_path is None:
                continue
            all_paths[os.path.join(root_path, media_hash.previous_path)] = os.path.join(root_path, media_hash.path)
        return all_paths

    def get_file_name(self):
        return os.path.basename(self.file_path)

    def get_root_path(self):
        return os.path.dirname(os.path.dirname(self.file_path))

    def generate_reference_hash(self):
        return hash_file(self.file_path, ascmhl_reference_hash_format)

    # build
    def append_hash(self, media_hash: MHLMediaHash):
        if media_hash.path == ".":
            self.process_info.root_media_hash = media_hash
        else:
            self.media_hashes.append(media_hash)
        self.media_hashes_path_map[media_hash.previous_path or media_hash.path] = media_hash
        self.media_hashes_path_map[media_hash.path] = media_hash

    def append_hash_list_reference(self, reference: MHLHashListReference):
        self.hash_list_references.append(reference)

    # log
    def log(self):
        logger.info("      filename: {0}".format(self.get_file_name()))
        logger.info("    generation: {0}".format(self.generation_number))

        self.creator_info.log()
        self.process_info.log()
        for media_hash in self.media_hashes:
            media_hash.log()


class MHLMediaHash:
    """
    class for representing one media hash for a (media) file
    managed by a MHLHashList object
    uses MHLHashEntry class for storing information about one hash value

    - public interface
        * accessing derived information

    - private interface
        * initialize new, empty MHLHashEntry for adding one hash value

    model member variables:
    hash_entries -- list of HashEntry objects to manage hash values (e.g. for different formats)

    attribute member variables:
    path -- relative file path to the file (supplements the root_path from the MHLHashList object)
    file_size -- size of the file
    last_modification_date -- last modification date as read from the filesystem
    previous_path -- file was renamed, new file path in path, old file path here

    other member variables:
    """

    hash_entries: List[MHLHashEntry]
    path: Optional[str]
    file_size: Optional[int]
    last_modification_date: Optional[datetime]
    is_directory: bool
    previous_path: Optional[str]

    # init
    def __init__(self):
        self.hash_entries = list()
        self.path = None
        self.file_size = None
        self.last_modification_date = None
        self.is_directory = False
        self.previous_path = None

    # methods to query for hashes
    def find_hash_entry_for_format(self, hash_format):
        for hash_entry in self.hash_entries:
            if hash_entry.hash_format == hash_format:
                return hash_entry
        return None

    # methods to build a media hash

    def append_hash_entry(self, hash_entry):
        hash_entry.media_hash = self
        self.hash_entries.append(hash_entry)

    # log

    def log(self):
        for hash_entry in self.hash_entries:
            self.log_hash_entry(hash_entry.hash_format)

    def log_hash_entry(self, hash_format):
        """find HashEntry object of a given format and print it"""
        for hash_entry in self.hash_entries:
            if hash_entry.hash_format != hash_format:
                continue
            indicator = " "
            if hash_entry.action == "failed":
                indicator = "!"
            elif self.is_directory:
                indicator = "d"

            hash_action = (hash_entry.action if hash_entry.action is not None else "").ljust(10)
            structure_string = ""
            if hash_entry.structure_hash_string is not None and hash_entry.structure_hash_string != "":
                structure_string = " (structure " + hash_entry.structure_hash_string + ")"
            action_string = ""
            if hash_action is not None and hash_action != "":
                action_string = "(action: " + hash_action + ")"
            logger.info(
                f"{indicator}"
                f" {hash_entry.hash_format.rjust(6)}"
                f": {hash_entry.hash_string.ljust(32)}"
                f"{structure_string}"
                f"{action_string}"
                f": {self.path}"
            )


class MHLHashEntry:
    """
    class to store one hash value
    managed by a MHLMediaHash object

    attribute member variables:
    hash_string -- string representation (hex) of the hash value
    hash_format -- string value, hash format, e.g. 'md5', 'xxh64'
    action -- action/result of verification, e.g. 'verified', 'failed', 'original'

    other member variables:
    """

    hash_string: str
    structure_hash_string: str
    hash_format: str
    hash_date: datetime
    action: Optional[str]

    def __init__(self, hash_format: str, hash_string: str, action: str = None, hash_date: datetime = None):
        self.hash_format = hash_format
        self.hash_string = hash_string
        self.structure_hash_string = None

        self.action = action
        if hash_date != None:
            self.hash_date = hash_date
        else:
            self.hash_date = datetime.now()


class MHLHashListReference:
    """
    class to store the ascmhlreference to a child history mhl file
    """

    path: Optional[str]
    reference_hash: Optional[str]

    def __init__(self):
        self.path = None
        self.reference_hash = None


class MHLCreatorInfo:
    """
    Stores the creator info that is part of the header of each hash list file
    """

    host_name: Optional[str]
    tool: Optional[MHLTool]
    creation_date: Optional[str]
    authors: List[MHLAuthor]
    location: Optional[str]
    comment: Optional[str]

    def __init__(self):
        self.host_name = None
        self.tool = None
        self.creation_date = None
        self.authors = []
        self.location = None
        self.comment = None

    def log(self):
        logger.info("      host_name: {0}".format(self.host_name))
        logger.info("           tool: {0} {1}".format(self.tool.name, self.tool.version))
        logger.info("  creation_date: {0}".format(self.creation_date))
        logger.info("       location: {0}".format(self.location))
        logger.info("        comment: {0}".format(self.comment))

    def summary(self):
        summary = ""
        if self.host_name is not None:
            summary += str(self.host_name)
        else:
            summary += "[unknown host]"
        if self.tool.name is not None:
            summary += ", " + str(self.tool.name)
            if self.tool.version is not None:
                summary += " " + str(self.tool.version)
        for author in self.authors:
            summary += ", " + str(author.name)
            if author.role is not None:
                summary += " [" + str(author.role) + "]"
            author_infos = []
            if author.email is not None:
                author_infos.append(author.email)
            if author.phone is not None:
                author_infos.append(author.phone)
            if len(author_infos) > 0:
                summary += " (" + ", ".join(author_infos) + ")"
        if self.location is not None:
            summary += ", at/in " + self.location
        if self.comment is not None:
            summary += ', "' + self.comment + '"'
        return summary


class MHLProcessInfo:
    """
    Stores the creator info that is part of the header of each hash list file
    """

    process: Optional[MHLProcess]
    root_media_hash: Optional[MHLMediaHash]
    ignore_spec: MHLIgnoreSpec
    hashlist_custom_basename: Optional[str]

    def __init__(self):
        self.process = None
        self.root_media_hash = None
        self.ignore_spec = MHLIgnoreSpec()
        self.hashlist_custom_basename = None

    def log(self):
        logger.info("        process: {0}".format(self.process))

    def summary(self):
        summary = ""
        if self.process is not None:
            summary += str(self.process)
        # TODO: ignore, root hash, ..
        return summary


class MHLTool:
    name: str
    version: str

    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version


class MHLProcess:
    process_type: str
    name: str

    def __init__(self, process_type: str, name: str = None):
        self.process_type = process_type
        self.name = name


class MHLAuthor:
    name: str
    email: Optional[str]
    phone: Optional[str]
    role: Optional[str]

    def __init__(self, name: str, email: str = None, phone: str = None, role: str = None):
        self.name = name
        self.email = email
        self.phone = phone
        self.role = role



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from .conftest import abspath_conversion_tests

from freezegun import freeze_time
from click.testing import CliRunner

from ascmhl.history import MHLHistory
import ascmhl.commands

scenario_output_path = "examples/scenarios/Output"
fake_ref_path = "/ref"

# def test_create_simple(fs, simple_mhl_history):
#    """
#    test that ensures that hasehs in a new format are also verified in child histories
#    used to verify fix of bug: https://github.com/ascmitc/mhl/issues/48
#    """
#
#    runner = CliRunner()
#    result = runner.invoke(ascmhl.commands.create, ['/root', '-h', 'md5'])
#    assert result.exit_code == 0


def test_create_nested(fs, nested_mhl_histories):
    """
    test that ensures that hasehs in a new format are also verified in child histories
    used to verify fix of bug: https://github.com/ascmitc/mhl/issues/48
    """

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "md5"])
    assert result.exit_code == 0



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
import shutil

from collections import defaultdict
from pathlib import PureWindowsPath
from typing import Dict, List

from . import chain_xml_parser, ignore
from . import logger
from .ignore import MHLIgnoreSpec
from .hashlist import MHLHashList, MHLHashEntry, MHLCreatorInfo, MHLProcessInfo
from .history import MHLHistory
from .utils import convert_posix_to_local_path


class MHLGenerationCreationSession:
    """
    class for representing a verification session

    the session is based on a MHLHistory object (that can be empty if no as-mhl folder exists yet).
    It is used to document the verification of single files and will handle verifying the given hash against
    existing generations. While being used by the verification process it will create a new generation (hash list)
    and write it to disk once the session is committed.

    - public interface
        * initialized with a MHLHistory object
        * adding hashes generated from certain files
        * committing of a session to write the new generation to disk
    """

    root_history: MHLHistory
    new_hash_lists: Dict[MHLHistory, MHLHashList]
    ignore_spec: MHLIgnoreSpec

    def __init__(self, history: MHLHistory, ignore_spec: MHLIgnoreSpec = MHLIgnoreSpec()):
        self.root_history = history
        self.new_hash_lists = defaultdict(MHLHashList)
        self.ignore_spec = ignore_spec

    def append_multiple_format_file_hashes(
        self, file_path, file_size, hash_lookup: Dict[str, str], file_modification_date, action=None, hash_date=None
    ) -> bool:
        """
        Adds file hashes to the history
        :param file_path: a string value representing the path to a file
        :param file_size: size of the file path in bytes
        :param hash_lookup: a dictionary of hash values keyed by the respective hash format
        :param file_modification_date: date the file was last modified
        :param action: a predetermined action for the entry.  defaults to none
        :param hash_date: date the hashes were generated
        :return a bool indicating if the hashes were successfully appended.  returns false if any failures occur
        """
        relative_path = self.root_history.get_relative_file_path(file_path)
        # TODO: handle if path is outside of history root path
        # Keep track of the number of failures
        failures = 0
        history, history_relative_path = self.root_history.find_history_for_path(relative_path)
        # for collections we cannot create a valid relative path (we are in the "wrong" history), but in that case
        # the file_path is inputted already as the relative path (a bit of implicit functionality here)
        if history_relative_path == None:
            history_relative_path = file_path

        # check if there is an existing hash in the other generations and verify
        original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

        hash_entries = [MHLHashEntry]
        # TODO: sort the format keys into a standard order for consistent output
        for hash_format, hash_string in hash_lookup.items():
            hash_entry = MHLHashEntry(hash_format, hash_string, hash_date=hash_date)
            if original_hash_entry is None:
                hash_entry.action = "original"
                logger.verbose(f"  created original hash for     {relative_path}  {hash_format}: {hash_string}")
            else:
                existing_hash_entry = history.find_first_hash_entry_for_path(history_relative_path, hash_format)
                if existing_hash_entry is not None:
                    if existing_hash_entry.hash_string == hash_string:
                        hash_entry.action = "verified"
                        logger.verbose(f"  verified                      {relative_path} {hash_format}: OK")
                    else:
                        hash_entry.action = "failed"
                        failures += 1
                        logger.error(
                            f"ERROR: hash mismatch for        {relative_path}  "
                            f"{hash_format} (old): {existing_hash_entry.hash_string}, "
                            f"{hash_format} (new): {hash_string}"
                        )
                else:
                    # in case there is no hash entry for this hash format yet
                    hash_entry.action = (  # mark as 'new' here, will be changed to verified in _validate_new_hash_list
                        "new"
                    )
                    logger.verbose(f"  created new (verif.) hash for {relative_path}  {hash_format}: {hash_string}")
            # collection behavior: overwrite action with action from flattened history
            if action != None:
                hash_entry.action = action

            # Add the generated entry to the list
            hash_entries.append(hash_entry)

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(
            history_relative_path, file_size, file_modification_date
        )

        # Add the new hash entries
        for hash_entry in hash_entries:
            media_hash.append_hash_entry(hash_entry)

        return failures == 0

    def append_file_hash(
        self, file_path, file_size, file_modification_date, hash_format, hash_string, action=None, hash_date=None
    ) -> bool:
        relative_path = self.root_history.get_relative_file_path(file_path)
        # TODO: handle if path is outside of history root path

        history, history_relative_path = self.root_history.find_history_for_path(relative_path)
        # for collections we cannot create a valid relative path (we are in the "wrong" history), but in that case
        # the file_path is inputted already as the relative path (a bit of implicit functionality here)
        if history_relative_path == None:
            history_relative_path = file_path

        # check if there is an existing hash in the other generations and verify
        original_hash_entry = history.find_original_hash_entry_for_path(history_relative_path)

        hash_entry = MHLHashEntry(hash_format, hash_string, hash_date=hash_date)
        if original_hash_entry is None:
            hash_entry.action = "original"
            if relative_path is not None:
                logger.verbose(f"  created original hash for     {relative_path}  {hash_format}: {hash_string}")
            else:
                # flattening works a bit different, because we don't add to individual (nested) histories
                logger.verbose(
                    f"  created original hash for     {convert_posix_to_local_path(file_path)}  {hash_format}: {hash_string}"
                )
        else:
            existing_hash_entry = history.find_first_hash_entry_for_path(history_relative_path, hash_format)
            if existing_hash_entry is not None:
                if existing_hash_entry.hash_string == hash_string:
                    hash_entry.action = "verified"
                    logger.verbose(f"  verified                      {relative_path}  {hash_format}: OK")
                else:
                    hash_entry.action = "failed"
                    logger.error(
                        f"ERROR: hash mismatch for        {relative_path}  "
                        f"{hash_format} (old): {existing_hash_entry.hash_string}, "
                        f"{hash_format} (new): {hash_string}"
                    )
            else:
                # in case there is no hash entry for this hash format yet
                hash_entry.action = "new"  # mark as 'new' here, will be changed to verified in _validate_new_hash_list
                logger.verbose(f"  created new (verif.) hash for {relative_path}  {hash_format}: {hash_string}")

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(
            history_relative_path, file_size, file_modification_date
        )

        # collection behavior: overwrite action with action from flattened history
        if action != None:
            hash_entry.action = action

        media_hash.append_hash_entry(hash_entry)
        return hash_entry.action != "failed"

    def append_multiple_format_directory_hashes(
        self, path, modification_date, content_hash_lookup: Dict[str, str], structure_hash_lookup: Dict[str, str]
    ) -> None:
        """
        Adds directory hashes to the history
        :param path: a string value representing the path to a file
        :param modification_date: date the file was last modified
        :param content_hash_lookup: a dictionary of content hash values keyed by the respective hash format
        :param structure_hash_lookup: a dictionary of structure hash values keyed by the respective hash format
        :return: none
        """
        relative_path = self.root_history.get_relative_file_path(path)
        # TODO: handle if path is outside of history root path

        history, history_relative_path = self.root_history.find_history_for_path(relative_path)

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(history_relative_path, None, modification_date)
        media_hash.is_directory = True

        # Add the content entries
        if content_hash_lookup:
            for hash_format, content_hash_string in content_hash_lookup.items():
                # Find the structure hash string
                structure_hash_string = structure_hash_lookup[hash_format]

                hash_entry = MHLHashEntry(hash_format, content_hash_string)
                # Attempt to add the structure, if available
                hash_entry.structure_hash_string = structure_hash_string
                media_hash.append_hash_entry(hash_entry)

                if relative_path == ".":
                    logger.verbose(
                        f"  calculated root hash  {hash_format}: "
                        f"{content_hash_string} (content), "
                        f"{structure_hash_string} (structure)"
                    )
                else:
                    logger.verbose(
                        f"  calculated directory hash for {relative_path}  {hash_format}: "
                        f"{content_hash_string} (content), "
                        f"{structure_hash_string} (structure)"
                    )
        else:
            logger.verbose(f"  added directory entry for     {relative_path}")

        # in case we just created the root media hash of the current hash list we also add it one history level above
        if new_hash_list.process_info.root_media_hash is media_hash and history.parent_history:
            parent_history = history.parent_history
            parent_relative_path = parent_history.get_relative_file_path(path)
            parent_hash_list = self.new_hash_lists[parent_history]
            parent_media_hash = parent_hash_list.find_or_create_media_hash_for_path(
                parent_relative_path, None, modification_date
            )
            parent_media_hash.is_directory = True
            if content_hash_lookup:
                for hash_format, content_hash_string in content_hash_lookup.items():
                    structure_hash_string = structure_hash_lookup[hash_format]
                    hash_entry = MHLHashEntry(hash_format, content_hash_string)
                    hash_entry.structure_hash_string = structure_hash_string
                    parent_media_hash.append_hash_entry(hash_entry)

    def append_directory_hashes(
        self, path, modification_date, hash_format, content_hash_string, structure_hash_string
    ) -> None:
        relative_path = self.root_history.get_relative_file_path(path)
        # TODO: handle if path is outside of history root path

        history, history_relative_path = self.root_history.find_history_for_path(relative_path)

        # in case the same file is hashes multiple times we want to add all hash entries
        new_hash_list = self.new_hash_lists[history]
        media_hash = new_hash_list.find_or_create_media_hash_for_path(history_relative_path, None, modification_date)
        media_hash.is_directory = True

        if content_hash_string:
            hash_entry = MHLHashEntry(hash_format, content_hash_string)
            hash_entry.structure_hash_string = structure_hash_string
            media_hash.append_hash_entry(hash_entry)
            if relative_path == ".":
                logger.verbose(
                    f"  calculated root hash  {hash_format}: "
                    f"{content_hash_string} (content), "
                    f"{structure_hash_string} (structure)"
                )
            else:
                logger.verbose(
                    f"  calculated directory hash for {relative_path}  {hash_format}: "
                    f"{content_hash_string} (content), "
                    f"{structure_hash_string} (structure)"
                )
        else:
            logger.verbose(f"  added directory entry for     {relative_path}")

        # in case we just created the root media hash of the current hash list we also add it one history level above
        if new_hash_list.process_info.root_media_hash is media_hash and history.parent_history:
            parent_history = history.parent_history
            parent_relative_path = parent_history.get_relative_file_path(path)
            parent_hash_list = self.new_hash_lists[parent_history]
            parent_media_hash = parent_hash_list.find_or_create_media_hash_for_path(
                parent_relative_path, None, modification_date
            )
            parent_media_hash.is_directory = True
            if content_hash_string:
                hash_entry = MHLHashEntry(hash_format, content_hash_string)
                hash_entry.structure_hash_string = structure_hash_string
                parent_media_hash.append_hash_entry(hash_entry)

    def commit(self, creator_info: MHLCreatorInfo, process_info: MHLProcessInfo, writeChain=True):
        """
        this method needs to create the generations of the children bottom up
        # so each history can reference the children correctly and can get the actual hash of the file
        """

        # store all references to child histories we will need when committing the parent history
        referenced_hash_lists: Dict[MHLHistory, List[MHLHashList]] = defaultdict(list)

        for history in MHLHistory.walk_child_histories(self.root_history):
            if history not in self.new_hash_lists:
                if history not in referenced_hash_lists:
                    continue
                new_hash_list = MHLHashList()
            else:
                new_hash_list = self.new_hash_lists[history]
            new_hash_list.referenced_hash_lists = referenced_hash_lists[history]
            new_hash_list.creator_info = creator_info

            # only for flattening we want to inject a custom root hash
            if not new_hash_list.process_info.root_media_hash:
                new_hash_list.process_info.root_media_hash = process_info.root_media_hash
            new_hash_list.process_info.hashlist_custom_basename = process_info.hashlist_custom_basename
            new_hash_list.process_info.process = process_info.process
            new_hash_list.process_info.ignore_spec = self.get_relevant_ignore_pattern(history)

            history.write_new_generation(new_hash_list)
            relative_generation_path = self.root_history.get_relative_file_path(new_hash_list.file_path)
            logger.verbose(f"Created new generation {relative_generation_path}")
            if history.parent_history is not None:
                referenced_hash_lists[history.parent_history].append(new_hash_list)

            if writeChain:
                # regular history ....
                chain_xml_parser.write_chain(history.chain, new_hash_list)
            else:
                # ... or flattened history manifest
                root_path = os.path.dirname(new_hash_list.file_path)
                if not os.path.exists(root_path):
                    logger.error(f"ERROR: folder {root_path} with flattened manifest does not exist")
                    return

                parent_folder = os.path.dirname(root_path)

                for file_name in os.listdir(root_path):
                    if file_name.endswith(".mhl"):
                        src_path = os.path.join(root_path, file_name)
                        dst_path = os.path.join(parent_folder, file_name)
                        shutil.move(src_path, dst_path)

                # Remove the folder if empty
                if not os.listdir(root_path):
                    os.rmdir(root_path)
                else:
                    logger.error(f"ERROR: temp folder not empty, did not remove {root_path}")

    def get_relevant_ignore_pattern(self, history) -> MHLIgnoreSpec:
        """
        Only store the relevant ignore patterns for an ascmhl-history and ignore others.
        This will split the pattern into the relevant bits for the lowest ascmhl-history relative to it
        """

        # get the ignore pattern from the latest history, if there is none, use the default pattern
        final_ignores = []
        latest_ignore_patterns = history.latest_ignore_patterns()
        if latest_ignore_patterns is None:
            final_ignores += ignore.default_ignore_list()
        else:
            final_ignores += latest_ignore_patterns

        ignore_patterns = self.ignore_spec.get_pattern_list()

        history_path = history.get_root_path()
        # get the highest parent history to build the correct relative paths for this generation
        if history.parent_history is not None:
            parent_history = history.parent_history
            while parent_history.parent_history is not None:
                parent_history = parent_history.parent_history

            parent_history_path = parent_history.get_root_path()
            parent_rel_path = os.path.relpath(history_path, parent_history_path)
            for pattern in ignore_patterns:
                if not pattern in final_ignores:
                    if "/" not in pattern:
                        final_ignores.append(pattern)
                    elif pattern.startswith("/**/"):
                        final_ignores.append(pattern)
                    elif belongs_to_child(pattern, history, parent_history_path):
                        # if child is ignored itself, we need to append the ignore pattern to the next parent
                        for child in history.walk_child_histories(history):
                            child_root: str
                            if os.name == "nt":
                                child_root = PureWindowsPath(child.get_root_path()).as_posix()
                            else:
                                child_root = child.get_root_path()
                            if child_root.endswith(pattern):
                                pattern = extract_ignore_pattern(pattern, history_path)
                                final_ignores.append(pattern)
                        continue
                    elif belongs_to_parent_or_neighbour(pattern, parent_rel_path):
                        continue
                    else:
                        pattern = extract_ignore_pattern(pattern, parent_rel_path)
                        final_ignores.append(pattern)
                else:
                    continue
        else:
            for pattern in ignore_patterns:
                if not pattern in final_ignores:
                    if (
                        not belongs_to_child(pattern, history, history_path)
                        and not pattern in ignore.default_ignore_list()
                    ):
                        if pattern.startswith("/"):
                            final_ignores.append(pattern)
                        elif pattern.find("/") != -1:
                            final_ignores.append(extract_ignore_pattern(pattern))
                        else:
                            final_ignores.append(pattern)
                    else:
                        for child in history.walk_child_histories(history):
                            child_root: str
                            if os.name == "nt":
                                child_root = PureWindowsPath(child.get_root_path()).as_posix()
                            else:
                                child_root = child.get_root_path()
                            if child_root.endswith(pattern):
                                if history == child.parent_history:
                                    pattern = extract_ignore_pattern(pattern)
                                    final_ignores.append(pattern)
        return MHLIgnoreSpec(final_ignores, latest_ignore_patterns)


def belongs_to_child(pattern, history, parent_history_path, ignore_child=None) -> bool:
    if pattern.startswith("/"):
        pattern = pattern[1:]
    for child in history.child_histories:
        if ignore_child == child:
            continue
        child_path = child.get_root_path()
        parent_rel_path = os.path.relpath(child_path, parent_history_path)
        if os.name == "nt":
            parent_rel_path = parent_rel_path.replace("\\", "/")
        if pattern.startswith(parent_rel_path):
            return True

    return False


def belongs_to_parent_or_neighbour(pattern, parent_rel_path) -> bool:
    if "/" not in pattern:
        return False
    if pattern.startswith("/") and "/" not in pattern[1:]:
        return True
    if pattern.endswith("/") and "/" not in pattern[:-1]:
        return False

    if pattern.startswith("**/"):
        return False

    pattern_parts = pattern.strip("/").split("/")
    parent_parts = parent_rel_path.strip(os.sep).split(os.sep)
    i = 0
    while i < min(len(pattern_parts), len(parent_parts)):
        if pattern_parts[i] != parent_parts[i]:
            if i > 0 and i < len(pattern_parts) - 1 and pattern_parts[i] == "**":
                return False
            return True
        else:
            i += 1
    if i == len(pattern_parts):
        return True

    return False


def extract_ignore_pattern(pattern: str, parent_rel_path=None) -> str:
    if pattern.startswith("/"):
        if "/" in pattern[1:]:
            pattern = pattern[1:]
        else:
            return pattern

    pattern_rel_path = _extract_pattern_relative_to_history(pattern, parent_rel_path)

    if pattern_rel_path is not None:
        if pattern.endswith("/"):
            if "/" in pattern[:-1]:
                return "/" + (pattern if pattern.startswith("**/") else pattern_rel_path)
            return pattern

        if pattern.endswith("/**"):
            if pattern.startswith("**/"):
                return pattern
            if "/" in pattern[:-3]:
                return pattern_rel_path if pattern.startswith("/") else "/" + pattern_rel_path
            return pattern

        if "/" in pattern[:-1]:
            return (
                "/" + pattern
                if pattern.startswith("**/")
                else pattern_rel_path if pattern_rel_path.startswith("/") else "/" + pattern_rel_path
            )

    return pattern


def _extract_pattern_relative_to_history(pattern: str, history_path=None) -> str:
    if pattern.startswith("**"):
        return pattern
    if history_path is None:
        return pattern

    pattern_parts = pattern.lstrip("/").split("/")
    history_path_parts = history_path.lstrip(os.sep).split(os.sep)

    i = j = k = 0

    while i < len(history_path_parts):
        if history_path_parts[i] == pattern_parts[0]:
            break
        else:
            i += 1

    while j < len(pattern_parts) and i + j < len(history_path_parts):
        if history_path_parts[i + j] == pattern_parts[j]:
            j += 1
        else:
            break

    result = ""

    while j < (len(pattern_parts)):
        if k == 0:
            result += pattern_parts[j]
            j += 1
            k += 1
        else:
            result += "/" + pattern_parts[j]
            j += 1
    if result != "":
        return result



# -- Path setup --------------------------------------------------------------

import os
import sys

# inserting root in path for autodoc
sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------

project = "ascmhl"
copyright = "2020, American Society of Cinematographers (ASC)"
author = "American Society of Cinematographers (ASC)"

master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_click",
    "sphinx_rtd_theme",
    "recommonmark",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
import sphinx_rtd_theme  # ignore your editor, this import is used

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from .conftest import abspath_conversion_tests
from .conftest import path_conversion_tests

from freezegun import freeze_time
from click.testing import CliRunner

from ascmhl.history import MHLHistory
import ascmhl.commands


@freeze_time("2020-01-16 09:15:00")
def test_simple_verify_fails_no_history(fs, simple_mhl_history):
    runner = CliRunner()
    os.rename("/root/ascmhl", "/root/_ascmhl")
    result = runner.invoke(ascmhl.commands.verify, ["-v", abspath_conversion_tests("/root/Stuff.txt")])
    assert result.exit_code == 30


@freeze_time("2020-01-16 09:15:00")
def test_simple_verify(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["-v", abspath_conversion_tests("/root")])
    assert (
        result.output
        == f"check folder at path: {abspath_conversion_tests('/root')}\nverification (xxh64) of file {str(path_conversion_tests('A/A1.txt'))}: OK\nverification (xxh64) of file"
        " Stuff.txt: OK\n"
    )
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_directory_verify(fs, simple_mhl_history):
    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.verify, ["-v", "-dh", abspath_conversion_tests("/root")])
    assert "verification of root folder   OK (generation 0001)\n" in result.output
    assert result.exit_code == 0


@freeze_time("2020-01-16 09:15:00")
def test_directory_verify_detect_changes(fs, simple_mhl_history):
    # add some more files and folders
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")
    os.mkdir("/root/emptyFolderB")
    os.mkdir("/root/emptyFolderC")
    os.mkdir("/root/emptyFolderC/emptyFolderCA")
    os.mkdir("/root/emptyFolderC/emptyFolderCB")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-v"])
    assert result.exit_code == 0

    result = runner.invoke(ascmhl.commands.verify, ["-v", "-dh", abspath_conversion_tests("/root")])
    # assert verification works before we purposefully mess things up
    assert result.exit_code == 0

    # altering the content of one file
    with open("/root/A/A2.txt", "a") as file:
        file.write("!!")

    result = runner.invoke(ascmhl.commands.verify, ["-v", "-dh", abspath_conversion_tests("/root")])
    assert "ERROR: content hash mismatch" in result.output
    assert result.exit_code == 12

    # rename one file
    os.rename("/root/B/B1.txt", "/root/B/B2.txt")

    result = runner.invoke(ascmhl.commands.verify, ["-v", "-dh", abspath_conversion_tests("/root")])
    assert "ERROR: structure hash mismatch" in result.output
    assert result.exit_code == 12


@freeze_time("2020-01-16 09:15:00")
def test_verify_single_file(fs, simple_mhl_history):
    runner = CliRunner()

    # verify relative path
    result = runner.invoke(
        ascmhl.commands.verify, ["-v", "-sf", str(path_conversion_tests("A/A1.txt")), abspath_conversion_tests("/root")]
    )
    assert result.exit_code == 0

    # verify absolute path
    result = runner.invoke(
        ascmhl.commands.verify,
        ["-v", "-sf", abspath_conversion_tests("/root/A/A1.txt"), abspath_conversion_tests("/root")],
    )
    assert result.exit_code == 0

    # verify existing, but not listed file
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    result = runner.invoke(
        ascmhl.commands.verify,
        ["-v", "-sf", str(path_conversion_tests("B/B1.txt")), abspath_conversion_tests("/root/")],
    )
    assert result.exit_code == 21

    # verify non existing file
    result = runner.invoke(
        ascmhl.commands.verify,
        ["-v", "-sf", str(path_conversion_tests("B/B2.txt")), abspath_conversion_tests("/root/")],
    )
    assert result.exit_code == 20

    # altering the content of one file
    with open("/root/A/A1.txt", "a") as file:
        file.write("!!")
    # verify
    result = runner.invoke(
        ascmhl.commands.verify, ["-v", "-sf", str(path_conversion_tests("A/A1.txt")), abspath_conversion_tests("/root")]
    )
    assert result.exit_code == 11



"""
__author__ = "Patrick Renner"
__copyright__ = "Copyright 2024, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

from .conftest import abspath_conversion_tests
from ascmhl.utils import convert_local_path_to_posix, convert_posix_to_local_path

from pathlib import PurePosixPath, PureWindowsPath
from .conftest import path_conversion_tests
import os


def test_conversion_to_posix(fs):
    windows_path = "\\foo\\bar\\test.txt"

    converted_path = PureWindowsPath(windows_path).as_posix()

    assert converted_path == "/foo/bar/test.txt"


def test_conversion_to_windows(fs):
    posix_path = "/foo/bar/test.txt"

    converted_path = str(PureWindowsPath(PurePosixPath(posix_path)))

    assert converted_path == "\\foo\\bar\\test.txt"


def test_conversion_from_posix_to_local():
    posix_path = "/foo/bar/test.txt"

    converted_path = str(path_conversion_tests(posix_path))
    if os.name == "posix":
        assert converted_path == posix_path
    elif os.name == "nt":
        assert converted_path == "\\foo\\bar\\test.txt"


def test_abspath_from_posix():
    # this test is only relevant on windows
    if os.name != "nt":
        return

    posix_path = "/foo/bar/test.txt"
    previous_cwd = os.getcwd()

    # the assumed drive depends on the working directory
    os.chdir("C:\\")
    converted_path = abspath_conversion_tests(posix_path)
    if os.name == "posix":
        assert converted_path == posix_path
    elif os.name == "nt":
        assert converted_path == "C:\\foo\\bar\\test.txt"

    os.chdir("D:\\")
    converted_path = abspath_conversion_tests(posix_path)
    if os.name == "posix":
        assert converted_path == posix_path
    elif os.name == "nt":
        assert converted_path == "D:\\foo\\bar\\test.txt"

    # restore the working directory to avoid influencing other tests
    os.chdir(previous_cwd)


def test_convert_to_xml_path():

    windows_path = "\\foo\\bar\\test.txt"
    posix_path = "/foo/bar/test.txt"

    if os.name == "posix":
        assert posix_path == convert_local_path_to_posix(posix_path)
    elif os.name == "nt":
        assert posix_path == convert_local_path_to_posix(windows_path)
    else:
        print(f"ERR: Unknown operating system: {os.name}")
        assert 0


def test_convert_xml_to_local_path():

    windows_path = "\\foo\\bar\\test.txt"
    posix_path = "/foo/bar/test.txt"

    if os.name == "posix":
        assert posix_path == convert_posix_to_local_path(posix_path)
    elif os.name == "nt":
        assert windows_path == convert_posix_to_local_path(posix_path)
    else:
        print(f"ERR: Unknown operating system: {os.name}")
        assert 0



"""
__author__ = "Katharina Böttcher"
__copyright__ = "Copyright 2024, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import os
from .conftest import abspath_conversion_tests
from .conftest import path_conversion_tests

from click.testing import CliRunner
from freezegun import freeze_time

import ascmhl


@freeze_time("2020-01-16 09:15:00")
def test_create_nested_succeed(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/C/C1.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/A/AA/ascmhl/0001_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/B/ascmhl/0001_B_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/B/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    with open("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert "0002_AA_2020-01-16_091500Z.mhl" in fileContents
        assert "B/ascmhl/0002_B_2020-01-16_091500Z.mhl" in fileContents
        assert "AA1.txt" not in fileContents
        assert "C1.txt" in fileContents

    fs.create_file("/root/A/AA/AA2.txt", contents="AA2\n")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/A/AA/ascmhl/0003_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl")
    assert not os.path.exists("/root/ascmhl/0003_B_2020-01-16_091500Z.mhl")
    with open("/root/ascmhl/0002_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert "0003_AA_2020-01-16_091500Z.mhl" in fileContents
        assert "AA2.txt" not in fileContents

    # test history command
    result = runner.invoke(ascmhl.commands.info, [abspath_conversion_tests("/root")])
    assert f"Child History at {abspath_conversion_tests('/root/A/AA')}:" in result.output
    assert result.output.count("Generation 3") == 2
    assert result.output.count("Generation 2") == 3


@freeze_time("2020-01-16 09:15:00")
def test_create_nested_mhl_file_modified(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/C/C1.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/A/AA/ascmhl/0001_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/B/ascmhl/0001_B_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/B/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    with open("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert "0002_AA_2020-01-16_091500Z.mhl" in fileContents
        assert "B/ascmhl/0002_B_2020-01-16_091500Z.mhl" in fileContents
        assert "AA1.txt" not in fileContents
        assert "C1.txt" in fileContents

    fs.create_file("/root/A/AA/AA2.txt", contents="AA2\n")

    with open("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl", "a") as mhl_file:
        mhl_file.write("changed content")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exception
    assert result.exit_code == 31


@freeze_time("2020-01-16 09:15:00")
def test_create_nested_mhl_file_missing(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/C/C1.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/A/AA/ascmhl/0001_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/B/ascmhl/0001_B_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/B/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    with open("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert "0002_AA_2020-01-16_091500Z.mhl" in fileContents
        assert "B/ascmhl/0002_B_2020-01-16_091500Z.mhl" in fileContents
        assert "AA1.txt" not in fileContents
        assert "C1.txt" in fileContents

    fs.create_file("/root/A/AA/AA2.txt", contents="AA2\n")

    os.remove("/root/A/AA/ascmhl/0001_AA_2020-01-16_091500Z.mhl")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exception
    assert result.exit_code == 33

    result = runner.invoke(ascmhl.commands.diff, [abspath_conversion_tests("/root")])
    assert result.exception
    assert result.exit_code == 33


@freeze_time("2020-01-16 09:15:00")
def test_create_nested_mhl_chain_missing(fs):
    fs.create_file("/root/Stuff.txt", contents="stuff\n")
    fs.create_file("/root/A/A1.txt", contents="A1\n")
    fs.create_file("/root/B/B1.txt", contents="B1\n")
    fs.create_file("/root/A/A2.txt", contents="A2\n")
    fs.create_file("/root/C/C1.txt", contents="A2\n")
    fs.create_file("/root/A/AA/AA1.txt", contents="AA1\n")
    os.mkdir("/root/emptyFolderA")

    runner = CliRunner()
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/A/AA"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/A/AA/ascmhl/0001_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root/B"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/B/ascmhl/0001_B_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/B/ascmhl/ascmhl_chain.xml")

    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert not result.exception
    assert os.path.exists("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/A/AA/ascmhl/0002_AA_2020-01-16_091500Z.mhl")
    assert os.path.exists("/root/ascmhl/ascmhl_chain.xml")
    with open("/root/ascmhl/0001_root_2020-01-16_091500Z.mhl", "r") as fin:
        fileContents = fin.read()
        assert "0002_AA_2020-01-16_091500Z.mhl" in fileContents
        assert "B/ascmhl/0002_B_2020-01-16_091500Z.mhl" in fileContents
        assert "AA1.txt" not in fileContents
        assert "C1.txt" in fileContents

    fs.create_file("/root/A/AA/AA2.txt", contents="AA2\n")
    os.remove("/root/A/AA/ascmhl/ascmhl_chain.xml")
    result = runner.invoke(ascmhl.commands.create, [abspath_conversion_tests("/root"), "-h", "xxh64", "-v"])
    assert result.exception
    assert result.exit_code == 32

    result = runner.invoke(ascmhl.commands.diff, [abspath_conversion_tests("/root")])
    assert result.exception
    assert result.exit_code == 32



# from click import get_current_context
# from . import ignore
#
# """
# functions in this module should only be invoked from the main thread
# get_current_context is not thread safe
# """
#
# IGNORE_PATTERN_LIST_KEY = f'{__name__}.IGNORE_PATTERN_LIST_KEY'
#
#
# def set_ignore_patterns(value):
#     ctx = get_current_context()
#     ctx.meta[IGNORE_PATTERN_LIST_KEY] = value
#
#
# def get_ignore_patterns():
#     return get_current_context().meta.get(IGNORE_PATTERN_LIST_KEY, ignore.default_ignore_list())



"""
__author__ = "Alexander Sahm"
__copyright__ = "Copyright 2020, Pomfort GmbH"

__license__ = "MIT"
__maintainer__ = "Patrick Renner, Alexander Sahm"
__email__ = "opensource@pomfort.com"
"""

import difflib
import filecmp
import glob
import os
import shutil
from importlib import reload
from .conftest import abspath_conversion_tests
from typing import List
from .conftest import path_conversion_tests

import pytest
from ascmhl import utils
from click.testing import CliRunner
from freezegun import freeze_time
from pyfakefs.fake_filesystem_unittest import Pause

import ascmhl.commands

scenario_output_path = "examples/scenarios/Output"
fake_ref_root_path = "/ref"


@pytest.fixture(autouse=True)
def version(monkeypatch):
    monkeypatch.setattr("ascmhl.__version__.ascmhl_tool_version", "0.3 alpha")
    monkeypatch.setattr("ascmhl.__version__.ascmhl_tool_name", "ascmhl.py")
    reload(ascmhl.commands)


@pytest.fixture()
def reference(fs):
    load_real_reference(fs)


@pytest.fixture
@freeze_time("2020-01-15 13:00:00")
def card_a002(fs):
    fs.create_file("/A002R2EC/Sidecar.txt", contents="BLOREM ipsum dolor sit amet, consetetur sadipscing elitr.\n")
    fs.create_file("/A002R2EC/Clips/A002C006_141024_R2EC.mov", contents="abcd\n")
    fs.create_file("/A002R2EC/Clips/A002C007_141024_R2EC.mov", contents="def\n")


@pytest.fixture
@freeze_time("2020-01-15 13:30:00")
def card_a003(fs):
    fs.create_file("/A003R2EC/Sidecar.txt", contents="Dolor sit amet, consetetur sadipscing elitr.\n")
    fs.create_file("/A003R2EC/Clips/A003C011_141024_R2EC.mov", contents="vbgh\n")
    fs.create_file("/A003R2EC/Clips/A003C012_141024_R2EC.mov", contents="zhgdr\n")


def load_real_reference(fake_fs):
    fake_fs.add_real_directory(scenario_output_path, target_path=fake_ref_root_path)


# custom dircmp to compare file contents as suggested in https://stackoverflow.com/a/24860799
class ContentDircmp(filecmp.dircmp):
    """
    Compare the content of dir1 and dir2. In contrast with filecmp.dircmp, this
    subclass compares the content of files with the same path.
    """

    # noinspection PyAttributeOutsideInit
    def phase3(self):
        """
        Find out differences between common files.
        Ensure we are using content comparison with shallow=False.
        """
        fcomp = filecmp.cmpfiles(self.left, self.right, self.common_files, shallow=False)
        self.same_files, self.diff_files, self.funny_files = fcomp


def dirs_are_equal(dir1, dir2):
    """
    Compare two directory trees content.
    Return False if they differ, True is they are the same.
    """
    result = True
    compared = ContentDircmp(dir1, dir2, ignore=[".DS_Store"])
    if compared.left_only or compared.right_only or compared.diff_files or compared.funny_files:
        compared.report_partial_closure()
        for file in compared.diff_files:
            print_diff_of_files(os.path.join(dir1, file), os.path.join(dir2, file))
        result = False
    for subdir in compared.common_dirs:
        if not dirs_are_equal(os.path.join(dir1, subdir), os.path.join(dir2, subdir)):
            result = False
    return result


def compare_dir_content(reference: str, dir_path: str) -> bool:
    if os.path.isabs(dir_path):
        relative_path = dir_path.lstrip("/")
    else:
        relative_path = dir_path
    ref_path = os.path.join(fake_ref_root_path, reference, relative_path)
    if os.path.isdir(dir_path) is True and os.path.isdir(ref_path) is True:
        result = dirs_are_equal(ref_path, dir_path)
    else:
        result = False

    return result


def compare_file_content(reference: str, path: str) -> bool:
    if os.path.isabs(path):
        relative_path = path.lstrip(os.sep)
    else:
        relative_path = path
    ref_path = os.path.join(fake_ref_root_path, reference, relative_path)
    if os.path.isfile(path) is True and os.path.isfile(ref_path) is True:
        result = filecmp.cmp(ref_path, path, shallow=False)
        if result is not True:
            print("\ngot different files:\n")
            print_diff_of_files(ref_path, path)
    else:
        result = False
    return result


def print_diff_of_files(path_a: str, path_b: str):
    with open(path_a, "r") as file:
        data_a = file.readlines()
    with open(path_b, "r") as file:
        data_b = file.readlines()
    diff = difflib.unified_diff(data_a, data_b, fromfile=path_a, tofile=path_b, n=2, lineterm="")
    for line in diff:
        print(line.rstrip())


def compare_files_against_reference(scenario_reference: str, folder_paths: List[str], fake_fs) -> bool:
    """
    checks if the scenario reference folder exists in the output folder if it doesn't we copy the folders there and
    consider it as reference. This way we can easily recreate the reference files of a scenario by deleting it on disk
    and running the tests
    """
    real_ref_path = os.path.join(scenario_output_path, scenario_reference)

    with Pause(fake_fs):
        if os.path.isdir(real_ref_path):
            # in case the reference path exists we compare the content
            compare_mode = True
        else:
            # otherwise we just write the result to the output folder
            # to be used as new reference for the next run
            compare_mode = False

    result = True
    for folder_path in folder_paths:
        validate_all_mhl_files_against_xml_schema(folder_path)
        assert os.path.isabs(folder_path)
        if compare_mode:
            result &= compare_dir_content(scenario_reference, folder_path)
        else:
            copy_fake_directory_to_real_fs(folder_path, real_ref_path, fake_fs)
            # also copy the log file
            with open("/log.txt", "rb") as file:
                data = file.read()
            with Pause(fake_fs):
                with open(os.path.join(real_ref_path, "log.txt"), "w+b") as dst_file:
                    dst_file.write(data)

    if compare_mode:
        # we always assume a log.txt to exists for each scenario, compare it as well to check differences in tool output
        result &= compare_file_content(scenario_reference, "/log.txt")
    return result


def validate_all_mhl_files_against_xml_schema(folder_path: str):
    """Find all mhl files created and validate them against the xsd"""
    mhl_files = glob.glob(abspath_conversion_tests(folder_path) + r"/**/*.mhl", recursive=True)
    runner = CliRunner()
    for file in mhl_files:
        result = runner.invoke(ascmhl.commands.xsd_schema_check, [file])
        assert result.exit_code == 0, result.output


def copy_fake_directory_to_real_fs(fake_dir: str, real_dir: str, fake_fs):
    """Utility function to copy a directory in the fake file system recursively to the real file system"""
    for root, _, file_names in os.walk(fake_dir):
        relative_root = root.lstrip(os.sep)
        current_real_dir = os.path.join(real_dir, relative_root)
        with Pause(fake_fs):
            os.makedirs(current_real_dir)
        for file_name in file_names:
            with open(os.path.join(root, file_name), "rb") as file:
                data = file.read()
            with Pause(fake_fs):
                with open(os.path.join(current_real_dir, file_name), "w+b") as dst_file:
                    dst_file.write(data)


def log_message(message: str):
    log_path = "/log.txt"
    with open(log_path, "a") as file:
        file.write(message)
        file.write("\n")


def execute_command(click_cmd, args):
    runner = CliRunner()
    log_message("")
    log_message(f'$ ascmhl.py {click_cmd.name} {" ".join(args)}')
    result = runner.invoke(click_cmd, args)
    log_message(result.output)
    log_message("")
    return result


@freeze_time("2020-01-16 09:15:00")
def test_scenario_01(fs, reference, card_a002):
    log_message("Scenario 01:")
    log_message("This is the most basic example. A camera card is copied to a travel drive and an ASC-MHL file is")
    log_message("created with hashes of all files on the card.")
    log_message("")

    log_message("Assume the source card /A002R2EC is copied to a travel drive /travel_01.")
    shutil.copytree("/A002R2EC", "/travel_01/A002R2EC")

    log_message("")
    log_message("Seal the copy on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/A002R2EC", "-h", "xxh64"])
    assert result.exit_code == 0
    assert compare_files_against_reference("scenario_01", ["/travel_01"], fs)


@freeze_time("2020-01-16 09:15:00")
def test_scenario_02(fs, reference, card_a002):
    log_message("Scenario 02:")
    log_message("In this scenario a copy is made, and then a copy of the copy. Two ASC-MHL are created during")
    log_message("this process, documenting the history of both copy processes.")
    log_message("")

    log_message("Assume the source card /A002R2EC is copied to a travel drive /travel_01.")
    shutil.copytree("/A002R2EC", "/travel_01/A002R2EC")

    log_message("")
    log_message("Seal the copy on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/A002R2EC", "-h", "xxh64"])
    assert result.exit_code == 0

    with freeze_time("2020-01-17 14:30:00"):
        log_message("")
        log_message("Assume the travel drive arrives at a facility on the next day.")
        log_message("And the folder A002R2EC is copied there from the travel drive to a file server at /file_server.")
        shutil.copytree("/travel_01/A002R2EC", "/file_server/A002R2EC")

        log_message("")
        log_message("Creating new generation for the folder A002R2EC again on the file server")
        log_message("this will verify all hashes, check for completeness and create a second generation")
        result = execute_command(ascmhl.commands.create, ["-v", "/file_server/A002R2EC", "-h", "xxh64"])
        assert result.exit_code == 0
        assert compare_files_against_reference("scenario_02", ["/travel_01", "/file_server"], fs)


@freeze_time("2020-01-16 09:15:00")
def test_scenario_03(fs, reference, card_a002):
    log_message("Scenario 03:")
    log_message("In this scenario the first hashes are created using the xxhash format. Different hash formats")
    log_message("might be required by systems used further down the workflow, so the second copy is verified")
    log_message("against the existing xxhash hashes, and additional MD5 hashes can be created and stored during")
    log_message("that process on demand.")
    log_message("")

    log_message("Assume the source card /A002R2EC is copied to a travel drive /travel_01.")
    shutil.copytree("/A002R2EC", "/travel_01/A002R2EC")

    log_message("")
    log_message("Seal the copy on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/A002R2EC", "-h", "xxh64"])
    assert result.exit_code == 0

    with freeze_time("2020-01-17 14:30:00"):
        log_message("")
        log_message("Assume the travel drive arrives at a facility on the next day.")
        log_message("And the folder A002R2EC is copied there from the travel drive to a file server at /file_server.")
        shutil.copytree("/travel_01/A002R2EC", "/file_server/A002R2EC")

        log_message("")
        log_message("Creating new generation for the folder A002R2EC again on the file server using MD5 hash format")
        log_message("this will verify all existing xxHashes, check for completeness,")
        log_message("and create a second generation with additional (new) MD5 hashes.")
        result = execute_command(ascmhl.commands.create, ["-v", "-h", "md5", "/file_server/A002R2EC"])
        assert result.exit_code == 0
        assert compare_files_against_reference("scenario_03", ["/travel_01", "/file_server"], fs)


@freeze_time("2020-01-16 09:15:00")
def test_scenario_04(fs, reference, card_a002):
    log_message("Scenario 04:")
    log_message("Copying a folder to a travel drive and from there to a file server with a hash mismatch in")
    log_message("one file.")
    log_message("")

    log_message("Assume the source card /A002R2EC is copied to a travel drive /travel_01.")
    shutil.copytree("/A002R2EC", "/travel_01/A002R2EC")

    log_message("")
    log_message("Seal the copy on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/A002R2EC", "-h", "xxh64"])
    assert result.exit_code == 0

    with freeze_time("2020-01-17 14:30:00"):
        log_message("")
        log_message("Assume the travel drive arrives at a facility on the next day.")
        log_message("And the folder A002R2EC is copied there from the travel drive to a file server at /file_server.")
        shutil.copytree("/travel_01/A002R2EC", "/file_server/A002R2EC")

        log_message("")
        log_message("afterwards we simulate that during the copy the Sidecar.txt got corrupt (altered")
        log_message("by modifying the file content")
        with open("/file_server/A002R2EC/Sidecar.txt", "a") as file:
            file.write("!!")

        log_message("")
        log_message("Creating new generation for the folder A002R2EC again on the file server.")
        log_message("This will verify all existing hashes and fail because Sidecar.txt was altered.")
        log_message("An error is shown and create a new generation that documents the failed verification")
        result = execute_command(ascmhl.commands.create, ["-v", "/file_server/A002R2EC", "-h", "xxh64"])
        assert result.exit_code == 11
        assert compare_files_against_reference("scenario_04", ["/travel_01", "/file_server"], fs)

        # the second generation will include the failed verification result


@freeze_time("2020-01-16 09:15:00")
def test_scenario_05(fs, reference, card_a002, card_a003):
    log_message("Scenario 05:")
    log_message("Copying two camera mags to a `Reels` folder on a travel drive, and the entire `Reels` folder")
    log_message("folder to a server.")
    log_message("")

    os.makedirs("/travel_01/Reels")
    os.makedirs("/file_server")

    log_message("Assume the source card /A002R2EC is copied to a Reels folder on travel drive /travel_01.")
    shutil.copytree("/A002R2EC", "/travel_01/Reels/A002R2EC")

    log_message("")
    log_message("Seal the copy of A002R2EC on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/Reels/A002R2EC", "-h", "xxh64"])
    assert result.exit_code == 0

    log_message("")
    log_message("Assume a second card /A003R2EC is copied to the same Reels folder on travel drive /travel_01.")
    shutil.copytree("/A003R2EC", "/travel_01/Reels/A003R2EC")

    log_message("")
    log_message("Seal the copy of A003R2EC on the travel drive /travel_01 to create the original mhl generation.")
    result = execute_command(ascmhl.commands.create, ["-v", "/travel_01/Reels/A003R2EC", "-h", "xxh64"])
    assert result.exit_code == 0

    with freeze_time("2020-01-17 14:30:00"):
        log_message("")
        log_message("Assume the travel drive arrives at a facility on the next day.")
        log_message("And the common Reels folder is copied from the travel drive to a file server at /file_server.")
        shutil.copytree("/travel_01/Reels", "/file_server/Reels")

        log_message("")
        log_message("Afterwards an arbitrary file Summary.txt is added to the Reels folder.")
        # An arbitrary file `Summary.txt` is added to the `Reels` folder.
        fs.create_file(
            "/file_server/Reels/Summary.txt", contents="This is a random summary\n\n* A002R2EC\n* A003R2EC\n"
        )

        log_message("")
        log_message("Creating new generation for the Reels folder on the file server.")
        log_message("This will verify all hashes, check for completeness and create two second generations")
        log_message("in the card sub folders A002R2EC and A003R2EC and an initial one for the Reels folder")
        log_message("with the original hash of the Summary.txt and references to the child histories")
        log_message("of the card sub folders.")
        # The `Reels` folder is verified on the file server.
        result = execute_command(ascmhl.commands.create, ["-v", "/file_server/Reels", "-h", "xxh64"])
        assert result.exit_code == 0
        assert compare_files_against_reference("scenario_05", ["/travel_01", "/file_server"], fs)
